import logging
from inspect import stack

import click

from com.enovation.helper.click.df_load import load_csv, load_xls
from com.enovation.helper.click.df_write import to_stdout, to_xlsx
from com.enovation.helper.click.df_modify import rename_columns, cleanse_null_values
from com.enovation.toolbox.predictability.click.dp_compute import dp_compute_date_predictability
from com.enovation.toolbox.predictability.click.dp_demo import dp_demo_date_predictability
from com.enovation.toolbox.predictability.click.dp_persist \
    import dp_write_bean_date_predictability, dp_load_bean_date_predictability
from com.enovation.toolbox.predictability.click.dp_graph import dp_graph_to_dash_date_predictability


_logger: logging.Logger = logging.getLogger(__name__)


@click.group(chain=True)
@click.option(
    '--verbose/--no-verbose', type=bool, default=None,
    help='Level of logging verbosity: INFO (--verbose), WARNING (default) or ERROR (--no-verbose).',
)
@click.pass_context
def enov(ctx_context, verbose):
    """
    The click application to call com.enovation modules...
    """

    if verbose is True:
        click.echo("logging: INFO")
        logging.basicConfig(level="INFO")
    elif verbose is False:
        click.echo("logging: ERROR")
        logging.basicConfig(level="ERROR")
    else:
        click.echo("logging: WARNING")
        logging.basicConfig(level="WARNING")

    _logger.debug(f"Function enov called.")
    _logger.info(f"Welcome!")

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    # Ensure that ctx_context.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    # This is effectively the context, that is shared across commands
    ctx_context.ensure_object(dict)

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


# com.enovation.helper.click.df_load
enov.add_command(load_csv)
enov.add_command(load_xls)

# com.enovation.helper.click.df_write
enov.add_command(to_stdout)
enov.add_command(to_xlsx)

# com.enovation.helper.click.df_modify
enov.add_command(rename_columns)
enov.add_command(cleanse_null_values)

# com.enovation.toolbox.predictability.click - Date Predictability
enov.add_command(dp_compute_date_predictability)
enov.add_command(dp_write_bean_date_predictability)
enov.add_command(dp_load_bean_date_predictability)
enov.add_command(dp_graph_to_dash_date_predictability)
enov.add_command(dp_demo_date_predictability)


if __name__ == '__main__':
    enov()



from datetime import datetime
from inspect import stack
from logging import Logger, getLogger

from enlighten import Manager, Counter
from numpy import where, ndarray, NaN
from pandas import DataFrame, MultiIndex, NaT

from com.enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
import enlighten

from com.enovation.toolbox.predictability.bean import PredictabilityBean


class DatePredictabilityComputer:
    _logger: Logger = getLogger(__name__)

    # Columns labels for the input file that is expected from the client
    str__input__column_label__key: str = "key"
    str__input__column_label__date: str = "date"
    str__input__column_label__measure: str = "measure"
    lst__input__column_labels: list = [
        str__input__column_label__key,
        str__input__column_label__date,
        str__input__column_label__measure,
    ]

    # Columns labels for the statistics
    str__statistics__column_label__measure_count: str = "measure_count"
    str__statistics__column_label__date_first: str = "date_first"
    str__statistics__column_label__date_last: str = "date_last"
    str__statistics__column_label__measure_first: str = "measure_first"
    str__statistics__column_label__measure_last: str = "measure_last"
    str__statistics__column_label__measure_min: str = "measure_min"
    str__statistics__column_label__measure_max: str = "measure_max"
    str__statistics__column_label__predictability_last: str = "predictability_last"

    lst__output__statistics_by_key__column_labels: list = [
        str__input__column_label__key,
        str__statistics__column_label__measure_count,
        str__statistics__column_label__date_first,
        str__statistics__column_label__date_last,
        str__statistics__column_label__measure_first,
        str__statistics__column_label__measure_last,
        str__statistics__column_label__measure_min,
        str__statistics__column_label__measure_max,
        str__statistics__column_label__predictability_last
    ]

    # Columns labels for working columns
    str__output__column_label__predictability: str = "predictability"

    # Default resampling frequency: 1D
    _str__default_resampling_rule: str = "1D"

    def compute_date_predictability(
            self,
            df_measures: DataFrame,
    ) -> PredictabilityBean:
        self._logger.debug(
            f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
            f"\n\t- df_measures: '{df_measures.shape}'"
        )

        # 1. We check structure
        self._check_structure(
            df_measures=df_measures,
            lst_column_labels=self.lst__input__column_labels,
        )
        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': structure checked.")

        # 2. We check and type the data
        the_df_typed = self._check_and_cleanse_data(
            df_measures=df_measures,
            str_column_label_key=self.str__input__column_label__key,
            str_column_label_date=self.str__input__column_label__date,
            str_column_label_measure=self.str__input__column_label__measure,
        )
        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': date checked and typed.")

        # 3. We resample
        the_df_resampled: DataFrame = self._resample(
            df_measures=the_df_typed,
            str_column_label_key=self.str__input__column_label__key,
            str_column_label_date=self.str__input__column_label__date,
            str_resample_rule=self._str__default_resampling_rule,
        )
        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': measures resampled.")

        # 4. We compute historical predictability
        the_df_historical_predictability: DataFrame = self._compute_historical_predictability(
            df_measures=the_df_resampled,
            str_column_label__key=self.str__input__column_label__key,
            str_column_label__date=self.str__input__column_label__date,
            str_column_label__measure=self.str__input__column_label__measure,
            str_column_label__predictability=self.str__output__column_label__predictability
        )
        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': predictability computed.")

        # 5. We instantiate return bean
        df_the_statistics_by_key: DataFrame = self._compute_statistics_by_key(
            df_measures=the_df_typed,
            df_historical_predictability=the_df_historical_predictability,

            str_column_label__key=self.str__input__column_label__key,
            str_column_label__date=self.str__input__column_label__date,
            str_column_label__measure=self.str__input__column_label__measure,

            str_column_label__predictability=self.str__output__column_label__predictability,

            str_column_label__measure_count=self.str__statistics__column_label__measure_count,
            str_column_label__date_first=self.str__statistics__column_label__date_first,
            str_column_label__date_last=self.str__statistics__column_label__date_last,
            str_column_label__measure_first=self.str__statistics__column_label__measure_first,
            str_column_label__measure_last=self.str__statistics__column_label__measure_last,
            str_column_label__measure_min=self.str__statistics__column_label__measure_min,
            str_column_label__measure_max=self.str__statistics__column_label__measure_max,
            str_column_label__predictability_last=self.str__statistics__column_label__predictability_last
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return PredictabilityBean(
            df_historical=the_df_historical_predictability,
            df_by_measure=df_measures,
            df_by_key=df_the_statistics_by_key,
        )

    def _check_structure(
            self,
            df_measures: DataFrame,
            lst_column_labels: list = None,
    ):
        """
        The measures provided by the client should comply to the following business rules:
            - BR_Structure_001: the measures do not have multiindex index
            - BR_Structure_002: the measures do not have multiindex columns
            - BR_Structure_003: the measures contains exactly 3 columns:
                - str__input__column_label__key: str = "key"
                - str__input__column_label__date: str = "date"
                - str__input__column_label__measure: str = "measure"

        The function:
            - Returns quietly in case the measures comply to these business rules
            - Or raises an alert in case the client's measures do not comply to any of these business rules.

        :param df_measures: the client's measures
        :param lst_column_labels: the expected columns labels
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        if lst_column_labels is None:
            lst_column_labels = self.lst__input__column_labels

        # BR_Structure_001: the measures do not have multiindex index
        if isinstance(df_measures.index, MultiIndex):
            raise Exception(f"BR_Structure_001 - Provided measures have a multiindex index, which is not expected. "
                            f"Flatten the index, and retry. Multiindex names: "
                            f"'{', '.join(df_measures.index.names)}'.")

        # BR_Structure_002: the measures do not have multiindex columns
        if isinstance(df_measures.columns, MultiIndex):
            raise Exception(f"BR_Structure_002 - Provided measures have multiindexed columns, which is not expected. "
                            f"Flatten the columns, and retry. Multiindex names: "
                            f"'{', '.join(df_measures.columns.names)}'.")

        # BR_Structure_003: the measures contains exactly 3 columns, aka "key", "date" and "measure"
        lst_the_missing_columns_in_df: list = list(
            set(lst_column_labels)
            - set(df_measures.columns)
        )
        if len(lst_the_missing_columns_in_df) > 0:
            raise Exception(f"BR_Structure_003 - Provided measures are missing the following mandatory columns: "
                            f"'{', '.join(lst_the_missing_columns_in_df)}'. Review your data, and retry.")
        lst_the_extra_columns_in_df: list = list(
            set(df_measures.columns)
            - set(lst_column_labels)
        )
        if len(lst_the_extra_columns_in_df) > 0:
            raise Exception(f"BR_Structure_003 - Provided measures contains extra unexpected columns: "
                            f"'{', '.join(lst_the_missing_columns_in_df)}'. Review your data, and retry.")

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def _check_and_cleanse_data(
            self,
            df_measures: DataFrame,
            str_column_label_key: str = str__input__column_label__key,
            str_column_label_date: str = str__input__column_label__date,
            str_column_label_measure: str = str__input__column_label__measure,
    ) -> DataFrame:
        """
        The measures provided by the client should comply to the following business rules:
            - BR_Data_001: at least 1 record is required in order to compute a predictability
            - BR_Data_002: there should not be any null value in the dataframe
            - BR_Data_003: both columns "date" and "measure" should be dates, or objects that can be typed to dates
            - BR_Data_004: there should not be any line with duplicated "date"

        Cleansing logic:
            - LOG_Data_001: Columns "date" and "measure" are typed into dates
            - LOG_Data_002: Remove useless measures, aka same measure duplicated across dates

        The function:
            - Returns quietly in case the client's measures comply to these business rules
            - Or raises an alert in case the client's measures do not comply to any of these business rules.

        :param df_measures: the client's measures
        :param str_column_label_date: the label of the column that bears the dates for the measures
        :param str_column_label_measure: the label of the column that bears the measures
        :return: the typed client's measures
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # BR_Data_001. At least one record
        if len(df_measures.index) == 0:
            raise Exception(f"BR_Data_001 - The parameter df_measures contains '0' records, while it should contain at "
                            f"least '1'.")

        # BR_Data_002. No null value
        if df_measures.isnull().values.any():
            raise Exception(f"BR_Data_002 - The parameter df_measures contains '"
                            f"{df_measures.isnull().values.sum()}' null values, which is not expected. Make "
                            f"sure you call the function with a dataframe df_measures which does not contain any null "
                            f"value.")

        # BR_Data_003. We retype both columns "date" and "measure" into datetime:
        # LOG_Data_001: Columns "date" and "measure" are typed into dates
        try:
            df_the_return: DataFrame = PandasDataframeTyper.type(
                df_to_type=df_measures.copy(),
                dict_columns_to_type={
                    str_column_label_date: "date",
                    str_column_label_measure: "date"
                },
                b_strict=False
            )
        except Exception as an_exception:
            raise Exception(
                f"BR_Data_003 - An exception occurred while typing columns '{str_column_label_date}' and "
                f"'{str_column_label_measure}' to dates. Check the data, and retry.") from an_exception

        # BR_Data_004. duplicated records for a given "date" (excluding the "measure" column)
        df_the_working_df: DataFrame = df_measures.drop(columns=[str_column_label_measure])
        df_the_working_df = df_the_working_df[df_the_working_df.duplicated()]
        if len(df_the_working_df.index) > 0:
            the_unique_duplicated_records: DataFrame = df_the_working_df.drop_duplicates()
            the_msg: str = f"BR_Data_004 - The parameter df_measures contains '{len(df_the_working_df.index)}' " \
                           f"records duplicated, which is not expected. Make sure your remove all duplicates for " \
                           f"values:\n\n{the_unique_duplicated_records.to_string()}\n\n"
            raise Exception(the_msg)

        # LOG_Data_002: Remove useless measures, aka same measure duplicated across dates
        df_the_return = df_the_return.sort_values(
            by=[
                str_column_label_key,
                str_column_label_date
            ]
        ).reset_index(drop=True)
        df_the_return["tmp_"+str_column_label_key] = df_the_return[str_column_label_key].shift(1)
        df_the_return["tmp_"+str_column_label_measure] = df_the_return[str_column_label_measure].shift(1)

        # We override the very last measure for each key, not to drop it in case it is similar to previous one
        # df_the_return = df_the_return.loc[-1, "tmp_"+str_column_label_key] = NaN
        # df_the_return.loc[-1, "tmp_"+str_column_label_measure] = NaT
        df_the_return.loc[
            df_the_return.groupby(str_column_label_key).tail(1).index,
            ["tmp_"+str_column_label_key, "tmp_"+str_column_label_measure]
        ] = [NaN, NaT]

        df_the_return = df_the_return[
            (df_the_return[str_column_label_key] != df_the_return["tmp_"+str_column_label_key])
            | (df_the_return[str_column_label_measure] != df_the_return["tmp_"+str_column_label_measure])
        ]
        df_the_return = df_the_return.drop(columns=[
            "tmp_" + str_column_label_key,
            "tmp_" + str_column_label_measure
        ])

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        return df_the_return

    def _resample(
            self,
            df_measures: DataFrame,
            str_column_label_key: str = str__input__column_label__key,
            str_column_label_date: str = str__input__column_label__date,
            str_resample_rule: str = _str__default_resampling_rule,
    ) -> DataFrame:

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        df_measures = df_measures\
            .set_index(str_column_label_date)\
            .groupby(str_column_label_key)\
            .resample(str_resample_rule)\
            .ffill()\
            .reset_index(level=str_column_label_date)\
            .reset_index(drop=True)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return df_measures

    def _compute_spot_predictability_for_a_given_key(
            self,
            df_measures: DataFrame,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,
    ) -> float:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        dt_the_last_date: datetime = df_measures[self.str__input__column_label__date].max()

        dt_the_last_measure: datetime = \
            df_measures[
                df_measures[self.str__input__column_label__date] == dt_the_last_date
            ][
                self.str__input__column_label__measure
            ].iloc[0]

        a_measures_predictabilities: ndarray = where(

            # If date is at last measure
            df_measures[str_column_label__date] == dt_the_last_measure,

            where(

                # If measure is spot on
                df_measures[str_column_label__measure] == dt_the_last_measure,
                1,
                0
            ),

            where(

                # If outside the cone
                abs((df_measures[str_column_label__measure]-dt_the_last_measure).dt.days)
                >= abs((df_measures[str_column_label__date]-dt_the_last_measure).dt.days),
                0,
                1
                - abs((df_measures[str_column_label__measure] - dt_the_last_measure).dt.days)
                / abs((df_measures[str_column_label__date] - dt_the_last_measure).dt.days)
            )
        )

        # Eventually, we compute the aggregated predictability
        f_the_return: float = \
            a_measures_predictabilities.sum() \
            / len(a_measures_predictabilities)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return f_the_return

    def _compute_historical_predictability(
            self,
            df_measures: DataFrame,
            str_column_label__key: str = str__input__column_label__key,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,
            str_column_label__predictability: str = str__output__column_label__predictability
    ) -> DataFrame:
        """
        A function to compute a date predictability across history.

        :param df_measures: measures for a given key
        :return: the historical predictability that would have been computed as of the last date.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # The expanding does not seem to work, raising an exception we did not succeed to fix:
        # "pandas.core.base.DataError: No numeric types to aggregate"
        # To workaround, we implement a loop...
        #
        # df_the_return: DataFrame = df_measures \
        #     .sort_values(by=[
        #         str_column_label__key,
        #         str_column_label__date
        #     ]) \
        #     .reset_index(drop=True)
        #
        # df_the_return[str_column_label__predictability] = df_the_return\
        #     .groupby(str_column_label__key)\
        #     .expanding(engine='numba', method='table')\
        #     .agg(
        #         self._compute_spot_predictability_for_a_given_key
        #     )

        df_measures[str_column_label__predictability] = -1

        # Setup progress bar
        obj_the_mgr: Manager = enlighten.get_manager()
        obj_the_counter: Counter = obj_the_mgr.counter(
            total=len(df_measures.index),
            desc='Progress:',
            unit='measures'
        )

        # For each line in the dataframe
        for t_i_row in df_measures.itertuples():

            # We compute the predictability:
            # - For the given key
            # - For all the previous records at the date
            f_i_predictability = self._compute_spot_predictability_for_a_given_key(
                df_measures=df_measures[
                    (df_measures[str_column_label__key] == getattr(t_i_row, str_column_label__key))
                    & (df_measures[str_column_label__date] <= getattr(t_i_row, str_column_label__date))
                ],
                str_column_label__date=str_column_label__date,
                str_column_label__measure=str_column_label__measure,
            )

            # We update the measures with these historical predictabilities
            df_measures.loc[getattr(t_i_row, 'Index'), str_column_label__predictability] = f_i_predictability

            # We update the progress bar
            obj_the_counter.update()

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return df_measures

    def _compute_statistics_by_key(
            self,
            df_measures: DataFrame,
            df_historical_predictability: DataFrame,

            str_column_label__key: str = str__input__column_label__key,
            str_column_label__date: str = str__input__column_label__date,
            str_column_label__measure: str = str__input__column_label__measure,

            str_column_label__predictability: str = str__output__column_label__predictability,

            str_column_label__measure_count: str = str__statistics__column_label__measure_count,
            str_column_label__date_first: str = str__statistics__column_label__date_first,
            str_column_label__date_last: str = str__statistics__column_label__date_last,
            str_column_label__measure_first: str = str__statistics__column_label__measure_first,
            str_column_label__measure_last: str = str__statistics__column_label__measure_last,
            str_column_label__measure_min: str = str__statistics__column_label__measure_min,
            str_column_label__measure_max: str = str__statistics__column_label__measure_max,
            str_column_label__predictability_last: str = str__statistics__column_label__predictability_last
    ) -> DataFrame:
        """
        todo.

        :param df_historical_predictability: measures for a given key
        :return: the historical predictability that would have been computed as of the last date.
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # 1. We compute global statistics for each predictability.V2.V2 key:
        #       - date_first
        #       - date_last
        #       - measure_min
        #       - measure_max
        df_the_stats_by_key: DataFrame = df_measures.groupby(str_column_label__key).agg({
            str_column_label__key: "count",
            str_column_label__date: ["min", "max"],
            str_column_label__measure: ["min", "max"],
        }).reset_index()
        df_the_stats_by_key.columns = [
            str_column_label__key,
            str_column_label__measure_count,
            str_column_label__date_first,
            str_column_label__date_last,
            # str_column_label__measure_first,
            # str_column_label__measure_last,
            str_column_label__measure_min,
            str_column_label__measure_max,
            # str__output__column_label__predictability
        ]

        # 2. We compute the 2 residual statistics from the measures depending on date first and last:
        #    - measure_first
        #    - measure_last
        df_the_stats_by_key = df_the_stats_by_key.merge(
            right=df_measures.rename(columns={
                str_column_label__date: str_column_label__date_first,
                str_column_label__measure: str_column_label__measure_first
            }),
            on=[str_column_label__key, str_column_label__date_first],
            how="left"
        )
        df_the_stats_by_key = df_the_stats_by_key.merge(
            right=df_measures.rename(columns={
                str_column_label__date: str_column_label__date_last,
                str_column_label__measure: str_column_label__measure_last
            }),
            on=[str_column_label__key, str_column_label__date_last],
            how="left"
        )

        # 3. We compute the statistic based on the historical predictabilities
        #    - predictability_last
        df_the_stats_by_key = df_the_stats_by_key.merge(
            right=df_historical_predictability[[
                str_column_label__key,
                str_column_label__date,
                str_column_label__predictability,
            ]].rename(columns={
                str_column_label__date: str_column_label__date_last,
                str_column_label__predictability: str_column_label__predictability_last
            }),
            on=[str_column_label__key, str_column_label__date_last],
            how="left"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return df_the_stats_by_key



import logging
from inspect import stack

import click
from click import BadParameter
from pandas import DataFrame

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dp_compute')
@click.pass_context
@click.argument('alias-in', type=str)
@click.option('-o', '--alias-out', help="Resulting dataframe.", type=str)
def dp_compute_date_predictability(ctx_context, alias_in, alias_out):
    """
    Rename dataframe columns.

    :param ctx_context: Click context.
    :param alias_in: dataframe to lad from the context, so its columns can be renamed.
    :param alias_out: alias to the cleansed dataframe, in case source dataframe should not be altered.
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"No object labelled '{alias_in}' could be found in context, which is not expected.",
            param_hint=f"Ensure you effectively have an object labelled '{alias_in}' in context.",
            ctx=ctx_context
        )
    elif not isinstance(ctx_context.obj[alias_in], DataFrame):
        raise BadParameter(
            message=f"Object labelled '{alias_in}' is of type '{type(ctx_context.obj[alias_in])}', which is not "
                    f"expected.",
            param_hint=f"Ensure to provide an 'DataFrame' instance.",
            ctx=ctx_context
        )

    if alias_out in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias_out}' already exists, of type "
            f"'{type(ctx_context.obj[alias_in])}'. This data will be overridden."
        )

    df_the_input: DataFrame = ctx_context.obj[alias_in]

    obj_the_computer: DatePredictabilityComputer = DatePredictabilityComputer()

    obj_the_bean: PredictabilityBean = obj_the_computer.compute_date_predictability(
        df_measures=df_the_input
    )

    # We refresh the context data store
    if alias_out:
        ctx_context.obj[alias_out] = obj_the_bean
    else:
        ctx_context.obj[alias_in] = obj_the_bean

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from datetime import timedelta, datetime
from inspect import stack
from logging import Logger, getLogger

from dash import Dash
from dash_html_components import Div, H3, Table, Tr, Td
from dash_core_components import Dropdown, Graph
from dash.dependencies import Input, Output
from pandas import DataFrame
from plotly.graph_objs import Figure, Scatter
from plotly.subplots import make_subplots

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


class DatePredictabilityGrapher:
    _logger: Logger = getLogger(__name__)

    def graph_predictability(
            self,
            obj_predictability: PredictabilityBean
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        obj_the_dash_instance: Dash = Dash(
            __name__,
            external_stylesheets=['https://codepen.io/chriddyp/pen/bWLwgP.css']
        )

        obj_the_dash_instance.layout = Div([

            # A title
            H3("Zoom on a given key"),

            # The filter to select a key (aka. a project, an opportunity, etc) to graph and zoom in
            Dropdown(
                id="key_drop_down",
                placeholder="Select a key to graph and zoom in...",
                options=[
                    {'label': x, 'value': x}
                    for x in obj_predictability.df_by_key[DatePredictabilityComputer.str__input__column_label__key]
                ],
            ),

            Div([
                # The graph
                Graph(id='predictability_graph'),

                # The table
                Table([
                    Tr([Td(x), Td("--", id=x)])
                    for x in obj_predictability.df_by_key.columns
                ]),
            ]),
        ])

        @obj_the_dash_instance.callback(
            [Output(x, 'children') for x in obj_predictability.df_by_key.columns],
            Input('key_drop_down', 'value'))
        def callback_update_table(str_key):

            # If we have a key selected in the drop down
            if str_key:

                # We return the statistics for this key, as a list
                return obj_predictability.df_by_key[
                           obj_predictability.df_by_key[
                               DatePredictabilityComputer.str__input__column_label__key
                           ] == str_key
                           ].iloc[0, :].tolist()

            # Else, we have no key selected in the drop down
            else:

                # We return "--" for each field
                return ["--"] * len(obj_predictability.df_by_key.columns)

        @obj_the_dash_instance.callback(
            Output('predictability_graph', 'figure'),
            Input('key_drop_down', 'value')
        )
        def callback_update_graph(str_key):

            # We instantiate the Figure object that will be rendered
            # fig_the_figure: Figure = Figure()
            fig_the_figure: Figure = make_subplots(specs=[[{"secondary_y": True}]])

            # Add figure title
            fig_the_figure.update_layout(
                title_text=f"Date Predictability for <b>{str_key}</b>"
            )

            # Set x-axis title
            fig_the_figure.update_xaxes(title_text=DatePredictabilityComputer.str__input__column_label__date)

            # Set y-axes titles
            fig_the_figure.update_yaxes(
                title_text=DatePredictabilityComputer.str__input__column_label__measure,
                secondary_y=False)
            fig_the_figure.update_yaxes(
                title_text=DatePredictabilityComputer.str__output__column_label__predictability,
                range=[0, 1],
                tickformat=',.0%',
                secondary_y=True)

            # If we have a key selected in the drop down
            if str_key:

                # We select the data for that selected key
                df_the_historical_data: DataFrame = obj_predictability.df_historical[
                    obj_predictability.df_historical[
                        DatePredictabilityComputer.str__input__column_label__key
                    ] == str_key
                    ]
                d_the_stats_data: dict = obj_predictability.df_by_key[
                                             obj_predictability.df_by_key[
                                                 DatePredictabilityComputer.str__input__column_label__key
                                             ] == str_key
                                             ].iloc[0, :].to_dict()

                # We trace the projected measures
                fig_the_figure.add_trace(
                    Scatter(
                        x=df_the_historical_data[DatePredictabilityComputer.str__input__column_label__date],
                        y=df_the_historical_data[DatePredictabilityComputer.str__input__column_label__measure],
                        name=DatePredictabilityComputer.str__input__column_label__measure,
                    )
                )

                # We trace the reference system: the cone converging towards the last measure
                dt_the_bottom_left_corner: datetime = min(
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last],
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__date_first]
                )
                dt_the_top_right_corner: datetime = max(
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last],
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__date_last]
                )
                dt_the_top_left_corner: datetime = max(
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last],
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last]
                    + timedelta(
                        days=(
                                d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last]
                                - d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__date_first]
                        ).days
                    )
                )
                dt_the_bottom_right_corner: datetime = min(
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last],
                    d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last]
                    - timedelta(
                        days=(
                                d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__date_last]
                                - d_the_stats_data[
                                    DatePredictabilityComputer.str__statistics__column_label__measure_last]
                        ).days
                    )
                )

                # 1. The last measure, as an horizontal line
                fig_the_figure.add_trace(
                    Scatter(
                        x=[
                            dt_the_bottom_left_corner,
                            dt_the_top_right_corner
                            # d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__date_first],
                            # d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__date_last],
                        ],
                        y=[
                            d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last],
                            d_the_stats_data[DatePredictabilityComputer.str__statistics__column_label__measure_last],
                        ],
                        name=DatePredictabilityComputer.str__statistics__column_label__measure_last,
                    )
                )

                # 2. The increasing cone
                fig_the_figure.add_trace(
                    Scatter(
                        x=[dt_the_bottom_left_corner, dt_the_top_right_corner],
                        y=[dt_the_bottom_left_corner, dt_the_top_right_corner],
                        name="Increasing cone",
                    )
                )

                # 3. The decreasing cone
                fig_the_figure.add_trace(
                    Scatter(
                        x=[dt_the_bottom_left_corner, dt_the_top_right_corner],
                        y=[dt_the_top_left_corner, dt_the_bottom_right_corner],
                        name="Decreasing cone",
                    )
                )

                # We trace the predictability
                fig_the_figure.add_trace(
                    Scatter(
                        x=df_the_historical_data[DatePredictabilityComputer.str__input__column_label__date],
                        y=df_the_historical_data[DatePredictabilityComputer.str__output__column_label__predictability],
                        name=DatePredictabilityComputer.str__output__column_label__predictability,
                    ),
                    secondary_y=True
                )

                fig_the_figure.update_traces(mode='markers+lines', hovertemplate=None)
                fig_the_figure.update_layout(legend=dict(y=0.5, traceorder='reversed', font_size=16))

                return fig_the_figure

            else:
                # Todo: return here a figure which is a rectangle with a message to select a key
                return Figure()

        obj_the_dash_instance.run_server(debug=True, use_reloader=False)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import json
from inspect import stack

import click
from click import BadParameter
from pandas import DataFrame
import logging

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('rename-columns')
@click.pass_context
@click.argument('alias-in', type=str)
@click.argument('dictionary', type=str)
@click.option('-o', '--alias-out', help="Resulting dataframe.", type=str)
def rename_columns(ctx_context, alias_in, dictionary, alias_out):
    """
    Rename dataframe columns.

    :param ctx_context: Click context.
    :param alias_in: dataframe to lad from the context, so its columns can be renamed.
    :param dictionary: dictionary to rename columns, in the form '{"col1":"new col1", ... , "colx":"new colx"}'.
    :param alias_out: alias to the cleansed dataframe, in case source dataframe should not be altered.
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot rename columns from '{alias_in}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively have a dataframe labelled '{alias_in}' loaded in context.",
            ctx=ctx_context
        )
    if alias_out in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias_out}' already exists, of type {type(alias_out)}. This data "
            f"will be overridden."
        )

    df_the_input: DataFrame = ctx_context.obj[alias_in]

    if not isinstance(df_the_input, DataFrame):
        raise BadParameter(
            message=f"Cannot rename columns from '{alias_in}', as this alias is of type '{type(df_the_input)}' while "
                    f"'DataFrame' is expected.",
            param_hint=f"Ensure you effectively assigned a DataFrame to the alias '{alias_in}'.",
            ctx=ctx_context
        )

    try:
        dict_the_dictionary: dict = json.loads(dictionary)
    except json.decoder.JSONDecodeError as error:
        raise BadParameter(
            message=f"Cannot transform '{dictionary}' as a dictionary to rename columns.",
            param_hint=f"Ensure you effectively provided a string with 'key':'value' assignment that can be used to "
                       f"rename columns.",
            ctx=ctx_context
        ) from error

    # We check the dictionary is a simple set of key:value assignments
    lst_the_errors: list = [
        "'"+k+"'" for k, v in dict_the_dictionary.items()
        if (
            (not isinstance(k, str))
            | (not isinstance(v, str))
        )
    ]
    if len(lst_the_errors):
        raise BadParameter(
            message=f"Cannot use '{dictionary}' as a dictionary to rename columns.",
            param_hint=f"Ensure you effectively provided a string with simple 'key as string':'value as string' "
                       f"assignment that can be used to rename columns.",
            ctx=ctx_context
        )

    # We check the dictionary only provides keys that effectively exist as columns
    lst_the_errors: list = list({k for k in dict_the_dictionary}-set(df_the_input.columns))
    if len(lst_the_errors) > 0:
        raise BadParameter(
            message=f"Columns '{', '.join(lst_the_errors)}' are to be renamed, but they do not exist in the dataframe.",
            param_hint=f"Ensure you effectively provide columns to rename that exist in the dataframe.",
            ctx=ctx_context
        )

    df_the_return: DataFrame = df_the_input.rename(columns=dict_the_dictionary)

    # We refresh the context data store
    if alias_out:
        ctx_context.obj[alias_out] = df_the_return
    else:
        ctx_context.obj[alias_in] = df_the_return

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('cleanse-null-values')
@click.pass_context
@click.argument('alias-in', type=str)
@click.argument('column', type=str)
@click.option('-v', '--value', help="Value to replace null.", type=str)
@click.option('-o', '--alias-out', help="Resulting dataframe.", type=str)
def cleanse_null_values(ctx_context, alias_in, column, value, alias_out):
    """
    Cleanses null values in listed columns, either replace with value when provided, or drop lines.

    :param alias_in: alias to a dataframe to cleanse.
    :param alias_out: alias to the cleansed dataframe, in case source dataframe should not be altered.
    :param column: columns labels to cleanse null values.
    :param value: value to replace null.
    :param ctx_context: Click context.
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias_in not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot cleanse '{alias_in}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively loaded some data, and labelled these with the alias '{alias_in}'.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias_in], DataFrame):
        raise BadParameter(
            message=f"Cannot cleanse '{alias_in}', as this alias is of type '{type(ctx_context.obj[alias_in])}' while "
                    f"'DataFrame' is expected.",
            param_hint=f"Ensure you effectively used operands to generate data as a DataFrame.",
            ctx=ctx_context
        )

    df_the_measures = ctx_context.obj[alias_in]

    if column not in df_the_measures.columns:
        raise Exception(f"The command cleanse_null_values cannot access column '{column}'. First, ensure this column "
                        f"effectively exists (aka. present among the columns of a loaded file, or present among "
                        f"generated data by a command), then call this function to cleanse null values.")

    if value:
        df_the_measures = df_the_measures[column].fillna(value)
    else:
        df_the_measures = df_the_measures[df_the_measures[column].notna()]

    # We refresh the context data store
    if alias_out:
        ctx_context.obj[alias_out] = df_the_measures
    else:
        ctx_context.obj[alias_in] = df_the_measures

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
from inspect import stack
from pathlib import Path

import click
from click import BadParameter
from pandas import DataFrame, ExcelWriter, read_excel

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_persister import DatePredictabilityPersister

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dp_persist')
@click.pass_context
@click.argument('alias', type=str)
@click.argument('file', type=click.Path(exists=False))
def dp_write_bean_date_predictability(ctx_context, alias, file):
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias not in ctx_context.obj:
        raise BadParameter(
            message=f"Cannot persist '{alias}', as this alias could not be found in context.",
            param_hint=f"Ensure you effectively computed a date predictability bean, and labelled it with the alias "
                       f"'{alias}'.",
            ctx=ctx_context
        )

    if not isinstance(ctx_context.obj[alias], PredictabilityBean):
        raise BadParameter(
            message=f"Cannot persist '{alias}', as this alias is of type '{type(ctx_context.obj[alias])}' while "
                    f"'PredictabilityBean' is expected.",
            param_hint=f"Ensure you effectively provide an instance of 'PredictabilityBean' to persist.",
            ctx=ctx_context
        )

    obj_the_data: PredictabilityBean = ctx_context.obj[alias]

    DatePredictabilityPersister().persist(
        obj_bean_to_persist=obj_the_data,
        p_file_path=Path(file),
    )

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('dp_load')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='csv')
def dp_load_bean_date_predictability(ctx_context, file, alias):
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with an instance of 'PredictabilityBean' from file '{file}'."
        )

    obj_the_bean: PredictabilityBean = DatePredictabilityPersister().load(
            p_file_path=Path(file),
    )

    # We update the context data store
    ctx_context.obj[alias] = obj_the_bean
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "load_predictability_bean"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
import os
from inspect import stack
from pathlib import Path

import click
from pandas import DataFrame

from com.enovation.toolbox.predictability.bean import PredictabilityBean
from com.enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dp_demo')
@click.argument('file', type=click.Path(exists=False))
@click.pass_context
def dp_demo_date_predictability(ctx_context, file):
    """
    Demonstration of how to use commands related to Date Predictability.

    FILE is the path to persist the outputs generated by the Date Predictability commands.
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    click.echo(f"**************************\nCopy past in your terminal the command generated below. The generated "
               f"command will do the following:"
               f"\n\t- Load a demo csv file"
               f"\n\t- Rename columns"
               f"\n\t- Compute date predictability")
    click.echo(
        f"python3 \"{Path(__file__).resolve().parents[5].joinpath('enov.py')}\" --verbose "
               
        f"load-csv -c 'Project and phase' -c 'Projection date' -c 'Projected ETA' "
        f"\"{Path(__file__).resolve().parents[0].joinpath('dp_demo.csv')}\" df_raw "
       
        f"rename-columns -o df_renamed df_raw "
        f"'{{\"Project and phase\":\"key\",\"Projection date\":\"date\",\"Projected ETA\":\"measure\"}}' "
        
        f"dp_compute -o df_predictability df_renamed "
        
        f"dp_persist df_predictability \"{file}\" "

        f"dp_graph df_predictability")

    """
    python3 ./enov.py --verbose \
      load-csv -c id -c timestamp -c ps_kick_off_date ~/Documents/07.PRO/04.ENOVATION/2021-06-29.MX.TestFiles/00.pipeline_history.csv df_raw \
      rename-columns -o df_renamed df_raw '{"id":"key","timestamp":"date","ps_kick_off_date":"measure"}' \
      cleanse-null-values -o df_cleansed df_renamed measure \
      compute_date_predictability -o df_predictability df_cleansed \
      write_predictability df_predictability ~/Documents/07.PRO/04.ENOVATION/2021-06-29.MX.TestFiles/jsg.bean.xlsx
      """

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
